@@ -45,21 +45,22 @@
 //! deserialization with the `serde` feature.
 
 #![allow(clippy::clippy::many_single_char_names)]
-#![doc(html_root_url = "https://docs.rs/versions/3.0.3")]
+#![warn(missing_docs)]
+#![doc(html_root_url = "https://docs.rs/versions/4.0.0")]
 
 use itertools::EitherOrBoth::{Both, Left, Right};
 use itertools::Itertools;
 use nom::branch::alt;
-use nom::bytes::complete::{tag, take_while1};
-use nom::character::complete::{alpha1, alphanumeric1, char, digit1};
-use nom::combinator::{map, map_res, opt, peek, recognize, value};
-use nom::multi::{many1, separated_list1};
+use nom::bytes::complete::tag;
+use nom::character::complete::{alphanumeric1, char, digit1};
+use nom::combinator::{fail, map, map_res, opt, peek, recognize, value};
+use nom::multi::separated_list1;
 use nom::IResult;
+use parsers::{alphanums, hyphenated_alphanums, unsigned};
 #[cfg(feature = "serde")]
 use serde::{Deserialize, Serialize};
 use std::cmp::Ordering;
 use std::cmp::Ordering::{Equal, Greater, Less};
-use std::fmt;
 use std::hash::{Hash, Hasher};
 
 mod parsers;
@@ -95,11 +96,14 @@ mod parsers;
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[derive(Debug, Eq, Clone, Default)]
 pub struct SemVer {
+    /// The major version.
     pub major: u32,
+    /// The minor version.
     pub minor: u32,
+    /// The patch version.
     pub patch: u32,
     /// `Some` implies that the inner `Vec` of the `Chunks` is not empty.
-    pub pre_rel: Option<Chunks>,
+    pub pre_rel: Option<Release>,
     /// `Some` implies that the inner `String` is not empty.
     pub meta: Option<String>,
 }
@@ -115,22 +119,19 @@ impl SemVer {
 
     /// A lossless conversion from `SemVer` to [`Version`].
     ///
-    /// **Note:** Unlike `SemVer`, `Version` expects its metadata before the
-    /// prerelease. For instance:
-    ///
     /// ```
     /// use versions::SemVer;
     ///
     /// let orig = "1.2.3-r1+git123";
     /// let ver = SemVer::new(orig).unwrap().to_version();
     ///
-    /// assert_eq!("1.2.3+git123-r1", format!("{}", ver));
+    /// assert_eq!("1.2.3-r1+git123", format!("{}", ver));
     /// ```
     pub fn to_version(&self) -> Version {
         let chunks = Chunks(vec![
-            Chunk(vec![Unit::Digits(self.major)]),
-            Chunk(vec![Unit::Digits(self.minor)]),
-            Chunk(vec![Unit::Digits(self.patch)]),
+            Chunk::Numeric(self.major),
+            Chunk::Numeric(self.minor),
+            Chunk::Numeric(self.patch),
         ]);
 
         Version {
@@ -158,7 +159,7 @@ impl SemVer {
             MChunk::Digits(self.patch, self.patch.to_string()),
         ];
         let next = self.pre_rel.as_ref().map(|pr| {
-            let chunks = pr.0.iter().filter_map(|c| c.mchunk()).collect();
+            let chunks = pr.0.iter().map(|c| c.mchunk()).collect();
             let next = self.meta.as_ref().map(|meta| {
                 let chunks = vec![MChunk::Plain(meta.clone())];
                 (Sep::Plus, Box::new(Mess { chunks, next: None }))
@@ -193,13 +194,16 @@ impl SemVer {
                         None => Greater,
                         Some(Greater) => Greater,
                         Some(Less) => Less,
-                        Some(Equal) => match other.chunks.0.get(3).as_ref().map(|c| c.0.as_slice())
-                        {
+                        // By this point, the major/minor/patch positions have
+                        // all been equal. If there is a fourth position, its
+                        // type, not its value, will determine which overall
+                        // version is greater.
+                        Some(Equal) => match other.chunks.0.get(3) {
                             // 1.2.3 > 1.2.3.git
-                            Some([Unit::Letters(_), ..]) => Greater,
+                            Some(Chunk::Alphanum(_)) => Greater,
                             // 1.2.3 < 1.2.3.0
-                            Some([Unit::Digits(_), ..]) => Less,
-                            Some([]) | None => self.pre_rel.cmp(&other.release),
+                            Some(Chunk::Numeric(_)) => Less,
+                            None => self.pre_rel.cmp(&other.release),
                         },
                     },
                 },
@@ -257,7 +261,7 @@ impl SemVer {
         let (i, minor) = parsers::unsigned(i)?;
         let (i, _) = char('.')(i)?;
         let (i, patch) = parsers::unsigned(i)?;
-        let (i, pre_rel) = opt(Chunks::pre_rel)(i)?;
+        let (i, pre_rel) = opt(Release::parse)(i)?;
         let (i, meta) = opt(parsers::meta)(i)?;
 
         let sv = SemVer {
@@ -321,8 +325,8 @@ impl Ord for SemVer {
     }
 }
 
-impl fmt::Display for SemVer {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+impl std::fmt::Display for SemVer {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "{}.{}.{}", self.major, self.minor, self.patch)?;
 
         if let Some(p) = &self.pre_rel {
@@ -367,10 +371,19 @@ impl fmt::Display for SemVer {
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[derive(Debug, PartialEq, Eq, Hash, Clone, Default)]
 pub struct Version {
+    /// An optional prefix that marks that some paradigm shift in versioning has
+    /// occurred between releases of some software.
     pub epoch: Option<u32>,
+    /// The main sections of the `Version`. Unlike [`SemVer`], these sections
+    /// are allowed to contain letters.
     pub chunks: Chunks,
+    /// This either indicates a prerelease like [`SemVer`], or a "release"
+    /// revision for software packages. In the latter case, a version like
+    /// `1.2.3-2` implies that the software itself hasn't changed, but that this
+    /// is the second bundling/release (etc.) of that particular package.
+    pub release: Option<Release>,
+    /// Some extra metadata that doesn't factor into comparison.
     pub meta: Option<String>,
-    pub release: Option<Chunks>,
 }
 
 impl Version {
@@ -425,9 +438,9 @@ impl Version {
 
     /// Convert to a `Mess` without considering the epoch.
     fn to_mess_continued(&self) -> Mess {
-        let chunks = self.chunks.0.iter().filter_map(|c| c.mchunk()).collect();
+        let chunks = self.chunks.0.iter().map(|c| c.mchunk()).collect();
         let next = self.release.as_ref().map(|cs| {
-            let chunks = cs.0.iter().filter_map(|c| c.mchunk()).collect();
+            let chunks = cs.0.iter().map(|c| c.mchunk()).collect();
             (Sep::Hyphen, Box::new(Mess { chunks, next: None }))
         });
         Mess { chunks, next }
@@ -484,8 +497,8 @@ impl Version {
     /// combination with other general `nom` parsers.
     pub fn parse(i: &str) -> IResult<&str, Version> {
         let (i, epoch) = opt(Version::epoch)(i)?;
-        let (i, chunks) = Chunks::parse(&Unit::string, i)?;
-        let (i, release) = opt(Chunks::pre_rel)(i)?;
+        let (i, chunks) = Chunks::parse(i)?;
+        let (i, release) = opt(Release::parse)(i)?;
         let (i, meta) = opt(parsers::meta)(i)?;
 
         let v = Version {
@@ -529,22 +542,22 @@ impl Ord for Version {
     }
 }
 
-impl fmt::Display for Version {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+impl std::fmt::Display for Version {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         if let Some(e) = self.epoch {
             write!(f, "{}:", e)?;
         }
 
         write!(f, "{}", self.chunks)?;
 
-        if let Some(m) = &self.meta {
-            write!(f, "+{}", m)?;
-        }
-
         if let Some(r) = &self.release {
             write!(f, "-{}", r)?;
         }
 
+        if let Some(m) = &self.meta {
+            write!(f, "+{}", m)?;
+        }
+
         Ok(())
     }
 }
@@ -558,7 +571,7 @@ impl fmt::Display for Version {
 /// With `Mess`, groups of letters/numbers are separated by a period, but can be
 /// further separated by the symbols `_-+:`.
 ///
-/// Unfortunately, [`Chunks`] cannot be used here, as some developers have
+/// Unfortunately, [`Chunk`] cannot be used here, as some developers have
 /// numbers like `1.003.04` which make parsers quite sad.
 ///
 /// Some `Mess` values have a shape that is tantalizingly close to a [`SemVer`].
@@ -587,7 +600,9 @@ impl fmt::Display for Version {
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[derive(Debug, PartialEq, Eq, Hash, Clone, Default)]
 pub struct Mess {
+    /// The first section of a `Mess`.
     pub chunks: Vec<MChunk>,
+    /// The rest of the `Mess`.
     pub next: Option<(Sep, Box<Mess>)>,
 }
 
@@ -621,7 +636,7 @@ impl Mess {
     /// Like [`Mess::nth`], but tries to parse out a full [`Chunk`] instead.
     fn nth_chunk(&self, x: usize) -> Option<Chunk> {
         let chunk = self.chunks.get(x)?.text();
-        let (i, c) = Chunk::parse(&Unit::string, chunk).ok()?;
+        let (i, c) = Chunk::parse_without_hyphens(chunk).ok()?;
         match i {
             "" => Some(c),
             _ => None,
@@ -680,8 +695,8 @@ impl Ord for Mess {
     }
 }
 
-impl fmt::Display for Mess {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+impl std::fmt::Display for Mess {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "{}", self.chunks.iter().join("."))?;
 
         if let Some((sep, m)) = &self.next {
@@ -769,8 +784,8 @@ impl Ord for MChunk {
     }
 }
 
-impl fmt::Display for MChunk {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+impl std::fmt::Display for MChunk {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         match self {
             MChunk::Digits(_, s) => write!(f, "{}", s),
             MChunk::Rev(_, s) => write!(f, "{}", s),
@@ -791,14 +806,18 @@ impl fmt::Display for MChunk {
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[derive(Debug, PartialEq, Eq, Hash, Clone)]
 pub enum Sep {
+    /// `:`
     Colon,
+    /// `-`
     Hyphen,
+    /// `+`
     Plus,
+    /// `_`
     Underscore,
 }
 
-impl fmt::Display for Sep {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+impl std::fmt::Display for Sep {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         let c = match self {
             Sep::Colon => ':',
             Sep::Hyphen => '-',
@@ -809,257 +828,79 @@ impl fmt::Display for Sep {
     }
 }
 
-/// A single unit of a version number. May be digits or a string of characters.
-///
-/// Groups of these are called `Chunk`s, and are the identifiers separated by
-/// periods in the source.
-///
-/// Please avoid using the `Unit::Letters` constructor yourself. Instead
-/// consider the [`Unit::from_string`] method to verify the input.
-#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
-#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone)]
-pub enum Unit {
-    Digits(u32),
-    Letters(String),
-}
-
-impl Unit {
-    // TODO This can be made simpler once `bool::then_some` is merged into
-    // stable Rust.
-    /// Smart constructor for a `Unit` made of letters.
-    pub fn from_string(s: String) -> Option<Unit> {
-        s.chars()
-            .all(|c| c.is_ascii_alphabetic())
-            .then(|| Unit::Letters(s))
-    }
-
-    fn parse<'a, 'b, F>(f: &'a F, i: &'b str) -> IResult<&'b str, Unit>
-    where
-        F: Fn(&'b str) -> IResult<&'b str, Unit>,
-    {
-        alt((Unit::digits, f))(i)
-    }
-
-    fn digits(i: &str) -> IResult<&str, Unit> {
-        parsers::unsigned(i).map(|(i, x)| (i, Unit::Digits(x)))
-    }
-
-    /// Parsing `Unit`s that belong to the main section of a `Version`.
-    fn string(i: &str) -> IResult<&str, Unit> {
-        alpha1(i).map(|(i, s)| (i, Unit::Letters(s.to_string())))
-    }
-
-    /// Parsing `Unit`s that belong to the prerelease section.
-    fn string_with_hyphens(i: &str) -> IResult<&str, Unit> {
-        map(
-            take_while1(|c: char| c.is_ascii_alphabetic() || c == '-'),
-            |s: &str| Unit::Letters(s.to_owned()),
-        )(i)
-    }
-
-    fn single_zero(i: &str) -> IResult<&str, Unit> {
-        map_res(tag("0"), |s: &str| s.parse::<u32>().map(Unit::Digits))(i)
-    }
-}
-
-impl fmt::Display for Unit {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        match self {
-            Unit::Digits(ns) => write!(f, "{}", ns),
-            Unit::Letters(cs) => write!(f, "{}", cs),
-        }
-    }
-}
-
-/// A logical unit of a version number.
-///
-/// Can consist of multiple letters and numbers. Groups of these (i.e.
-/// [`Chunks`]) are separated by periods to form a full version number.
-///
-/// Defined as a newtype wrapper so that we can define custom parser and trait
-/// methods.
-///
-/// # Examples
-///
-/// - `r3`
-/// - `0rc1`
-/// - `20150826`
+/// [`Chunk`]s that have comparison behaviour according to SemVer's rules for
+/// prereleases.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[derive(Debug, PartialEq, Eq, Hash, Clone)]
-pub struct Chunk(pub Vec<Unit>);
+pub struct Release(pub Vec<Chunk>);
 
-impl Chunk {
-    /// If this `Chunk` is made up of a single [`Unit::Digits`].
-    /// then pull out the inner value.
-    ///
-    /// ```
-    /// use versions::{Chunk, Unit};
-    ///
-    /// let v = Chunk(vec![Unit::Digits(1)]);
-    /// assert_eq!(Some(1), v.single_digit());
-    ///
-    /// let v = Chunk(vec![Unit::Letters("abc".to_string())]);
-    /// assert_eq!(None, v.single_digit());
-    ///
-    /// let v = Chunk(vec![Unit::Digits(1), Unit::Letters("abc".to_string())]);
-    /// assert_eq!(None, v.single_digit());
-    /// ```
-    pub fn single_digit(&self) -> Option<u32> {
-        match self.0.as_slice() {
-            [Unit::Digits(n)] => Some(*n),
-            _ => None,
-        }
-    }
-
-    /// Like [`Chunk::single_digit`], but will grab the `u32` even if there were
-    /// more values in the `Chunk`.
-    ///
-    /// ```
-    /// use versions::{Chunk, Unit};
-    ///
-    /// let v = Chunk(vec![Unit::Digits(1)]);
-    /// assert_eq!(Some(1), v.single_digit_lenient());
-    ///
-    /// let v = Chunk(vec![Unit::Letters("abc".to_string())]);
-    /// assert_eq!(None, v.single_digit_lenient());
-    ///
-    /// let v = Chunk(vec![Unit::Digits(0), Unit::Letters("a".to_string())]);
-    /// assert_eq!(Some(0), v.single_digit_lenient());
-    /// ```
-    pub fn single_digit_lenient(&self) -> Option<u32> {
-        match self.0.first() {
-            Some(Unit::Digits(n)) => Some(*n),
-            _ => None,
-        }
-    }
-
-    // In `Option` since there's no way to guarantee that the `Vec` isn't empty.
-    // We could use `nonempty::NonEmpty`, but that has pains in creating the
-    // `NonEmpty`, and also expands the API outside of std collections.
-    fn mchunk(&self) -> Option<MChunk> {
-        match self.0.as_slice() {
-            [] => None,
-            [Unit::Digits(u)] => Some(MChunk::Digits(*u, u.to_string())),
-            [Unit::Letters(s), Unit::Digits(u)] if s == "r" => {
-                Some(MChunk::Rev(*u, format!("r{}", u)))
-            }
-            [Unit::Letters(s)] => Some(MChunk::Plain(s.clone())),
-            _ => Some(MChunk::Plain(format!("{}", self))),
-        }
-    }
-
-    /// Does this `Chunk` start with a letter?
-    fn is_lettered(&self) -> bool {
-        matches!(self.0.as_slice(), [Unit::Letters(_), ..])
-    }
-
-    fn parse<'a, 'b, F>(f: &'a F, i: &'b str) -> IResult<&'b str, Chunk>
-    where
-        F: Fn(&'b str) -> IResult<&'b str, Unit>,
-    {
-        Chunk::units(f, i).map(|(i, v)| (i, Chunk(v)))
-    }
-
-    /// Handling `0` is a bit tricky. We can't allow runs of zeros in a chunk,
-    /// since a version like `1.000.1` would parse as `1.0.1`.
-    fn units<'a, 'b, F>(f: &'a F, i: &'b str) -> IResult<&'b str, Vec<Unit>>
-    where
-        F: Fn(&'b str) -> IResult<&'b str, Unit>,
-    {
-        alt((
-            |i| Chunk::zero_with_letters(f, i),
-            map(Unit::single_zero, |u| vec![u]),
-            many1(|i| Unit::parse(f, i)),
-        ))(i)
-    }
-
-    fn zero_with_letters<'a, 'b, F>(f: &'a F, i: &'b str) -> IResult<&'b str, Vec<Unit>>
-    where
-        F: Fn(&'b str) -> IResult<&'b str, Unit>,
-    {
-        let (i, z) = Unit::single_zero(i)?;
-        let (i, s) = Unit::string(i)?;
-        let (i, c) = opt(|i| Chunk::units(f, i))(i)?;
-
-        let mut us = vec![z, s];
-        if let Some(x) = c {
-            us.extend(x);
-        }
-
-        Ok((i, us))
+impl Release {
+    fn parse(i: &str) -> IResult<&str, Release> {
+        let (i, _) = char('-')(i)?;
+        map(separated_list1(char('.'), Chunk::parse), Release)(i)
     }
 }
 
-impl PartialOrd for Chunk {
+impl PartialOrd for Release {
     fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
         Some(self.cmp(other))
     }
 }
 
-/// A custom implementation.
-impl Ord for Chunk {
+impl Ord for Release {
     fn cmp(&self, other: &Self) -> Ordering {
         self.0
             .iter()
             .zip_longest(&other.0)
             .find_map(|eob| match eob {
-                // Different from the `Ord` instance of `Chunks`, if we've
-                // iterated this far and one side has fewer chunks, it must be
-                // the "greater" version. A Chunk break only occurs in a switch
-                // from digits to letters and vice versa, so anything "extra"
-                // must be an `rc` marking or similar. Consider `1.1` compared
-                // to `1.1rc1`.
-                Left(_) => Some(Less),
-                Right(_) => Some(Greater),
-                // The usual case where the `Unit` types match, as in `1.2.3.4`
-                // vs `1.2.4.0`.
-                Both(Unit::Digits(a), Unit::Digits(b)) => match a.cmp(b) {
+                Both(a, b) => match a.cmp_semver(&b) {
+                    Less => Some(Less),
+                    Greater => Some(Greater),
                     Equal => None,
-                    ord => Some(ord),
                 },
-                Both(Unit::Letters(a), Unit::Letters(b)) => match a.cmp(b) {
-                    Equal => None,
-                    ord => Some(ord),
-                },
-                // The arbitrary decision to prioritize Letters over Digits has
-                // the effect of allowing this instance to work for `SemVer` as
-                // well as `Version`.
-                Both(Unit::Digits(_), Unit::Letters(_)) => Some(Less),
-                Both(Unit::Letters(_), Unit::Digits(_)) => Some(Greater),
+                // From the Semver spec: A larger set of pre-release fields has
+                // a higher precedence than a smaller set, if all the preceding
+                // identifiers are equal.
+                Left(_) => Some(Greater),
+                Right(_) => Some(Less),
             })
             .unwrap_or(Equal)
     }
 }
 
-impl fmt::Display for Chunk {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        for u in &self.0 {
-            write!(f, "{}", u)?;
+impl std::fmt::Display for Release {
+    // FIXME Fri Jan  7 11:44:50 2022
+    //
+    // Use `itersperse` here once it stabilises.
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        match self.0.as_slice() {
+            [] => Ok(()),
+            [c] => write!(f, "{}", c),
+            [c, rest @ ..] => {
+                write!(f, "{}", c)?;
+
+                for r in rest {
+                    write!(f, ".{}", r)?;
+                }
+
+                Ok(())
+            }
         }
-        Ok(())
     }
 }
 
-/// Multiple [`Chunk`] values.
-///
-/// Defined as a newtype wrapper so that we can define custom parser and trait
-/// methods.
+/// [`Chunk`]s that have a comparison behaviour specific to [`Version`].
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[derive(Debug, PartialEq, Eq, Hash, Clone, Default)]
 pub struct Chunks(pub Vec<Chunk>);
 
 impl Chunks {
-    fn parse<'a, 'b, F>(f: &'a F, i: &'b str) -> IResult<&'b str, Chunks>
-    where
-        F: Fn(&'b str) -> IResult<&'b str, Unit>,
-    {
-        let (i, cs) = separated_list1(char('.'), |i| Chunk::parse(f, i))(i)?;
-        Ok((i, Chunks(cs)))
-    }
-
-    fn pre_rel(i: &str) -> IResult<&str, Chunks> {
-        let (i, _) = char('-')(i)?;
-        Chunks::parse(&Unit::string_with_hyphens, i)
+    // Intended for parsing a `Version`.
+    fn parse(i: &str) -> IResult<&str, Chunks> {
+        map(
+            separated_list1(char('.'), Chunk::parse_without_hyphens),
+            Chunks,
+        )(i)
     }
 }
 
@@ -1075,30 +916,208 @@ impl Ord for Chunks {
             .iter()
             .zip_longest(&other.0)
             .find_map(|eob| match eob {
-                // If all chunks up until this point were equal, but one side
-                // continues on with "lettered" sections, these are considered
-                // to be indicating a beta/prerelease, and thus are *less* than
-                // the side who already ran out of chunks.
-                Left(c) if c.is_lettered() => Some(Less),
-                Right(c) if c.is_lettered() => Some(Greater),
-                // If one side has run out of chunks to compare but the other
-                // hasn't, the other must be newer.
+                Both(a, b) => match a.cmp_lenient(&b) {
+                    Less => Some(Less),
+                    Greater => Some(Greater),
+                    Equal => None,
+                },
+                // From the Semver spec: A larger set of pre-release fields has
+                // a higher precedence than a smaller set, if all the preceding
+                // identifiers are equal.
                 Left(_) => Some(Greater),
                 Right(_) => Some(Less),
-                // The usual case.
-                Both(a, b) => match a.cmp(b) {
-                    Equal => None,
-                    ord => Some(ord),
-                },
             })
             .unwrap_or(Equal)
     }
 }
 
-impl fmt::Display for Chunks {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        let s: String = self.0.iter().map(|c| format!("{}", c)).join(".");
-        write!(f, "{}", s)
+impl std::fmt::Display for Chunks {
+    // FIXME Fri Jan  7 11:44:50 2022
+    //
+    // Use `itersperse` here once it stabilises.
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        match self.0.as_slice() {
+            [] => Ok(()),
+            [c] => write!(f, "{}", c),
+            [c, rest @ ..] => {
+                write!(f, "{}", c)?;
+
+                for r in rest {
+                    write!(f, ".{}", r)?;
+                }
+
+                Ok(())
+            }
+        }
+    }
+}
+
+/// A logical unit of a version number.
+///
+/// Either entirely numerical (with no leading zeroes) or entirely
+/// alphanumerical (with a free mixture of numbers, letters, and hyphens).
+///
+/// Groups of these (like [`Release`]) are separated by periods to form a full
+/// section of a version number.
+///
+/// # Examples
+///
+/// - `1`
+/// - `20150826`
+/// - `r3`
+/// - `0rc1-abc3`
+#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[derive(Debug, PartialEq, Eq, Hash, Clone)]
+pub enum Chunk {
+    /// A nice, pure number.
+    Numeric(u32),
+    /// A mixture of letters, numbers, and hyphens.
+    Alphanum(String),
+}
+
+impl Chunk {
+    /// If this `Chunk` is made up of a single digit, then pull out the inner
+    /// value.
+    ///
+    /// ```
+    /// use versions::Chunk;
+    ///
+    /// let v = Chunk::Numeric(1);
+    /// assert_eq!(Some(1), v.single_digit());
+    ///
+    /// let v = Chunk::Alphanum("abc".to_string());
+    /// assert_eq!(None, v.single_digit());
+    ///
+    /// let v = Chunk::Alphanum("1abc".to_string());
+    /// assert_eq!(None, v.single_digit());
+    /// ```
+    pub fn single_digit(&self) -> Option<u32> {
+        match self {
+            Chunk::Numeric(n) => Some(*n),
+            Chunk::Alphanum(_) => None,
+        }
+    }
+
+    /// Like [`Chunk::single_digit`], but will grab a leading `u32` even if
+    /// followed by letters.
+    ///
+    /// ```
+    /// use versions::Chunk;
+    ///
+    /// let v = Chunk::Numeric(1);
+    /// assert_eq!(Some(1), v.single_digit_lenient());
+    ///
+    /// let v = Chunk::Alphanum("abc".to_string());
+    /// assert_eq!(None, v.single_digit_lenient());
+    ///
+    /// let v = Chunk::Alphanum("1abc".to_string());
+    /// assert_eq!(Some(1), v.single_digit_lenient());
+    /// ```
+    pub fn single_digit_lenient(&self) -> Option<u32> {
+        match self {
+            Chunk::Numeric(n) => Some(*n),
+            Chunk::Alphanum(s) => unsigned(s).ok().map(|(_, n)| n),
+        }
+    }
+
+    fn parse(i: &str) -> IResult<&str, Chunk> {
+        alt((Chunk::alphanum, Chunk::numeric))(i)
+    }
+
+    fn parse_without_hyphens(i: &str) -> IResult<&str, Chunk> {
+        alt((Chunk::alphanum_without_hyphens, Chunk::numeric))(i)
+    }
+
+    // A clever interpretation of the grammar of "alphanumeric identifier".
+    // Instead of having a big, composed parser that structurally accounts for
+    // the presence of a "non-digit", we just check for one after the fact.
+    fn alphanum(i: &str) -> IResult<&str, Chunk> {
+        let (i2, ids) = hyphenated_alphanums(i)?;
+
+        if ids.contains(|c: char| c.is_ascii_alphabetic() || c == '-') {
+            Ok((i2, Chunk::Alphanum(ids.to_string())))
+        } else {
+            fail(i)
+        }
+    }
+
+    fn alphanum_without_hyphens(i: &str) -> IResult<&str, Chunk> {
+        let (i2, ids) = alphanums(i)?;
+
+        if ids.contains(|c: char| c.is_ascii_alphabetic()) {
+            Ok((i2, Chunk::Alphanum(ids.to_string())))
+        } else {
+            fail(i)
+        }
+    }
+
+    fn numeric(i: &str) -> IResult<&str, Chunk> {
+        map(unsigned, Chunk::Numeric)(i)
+    }
+
+    fn mchunk(&self) -> MChunk {
+        // FIXME Fri Jan  7 12:34:24 2022
+        //
+        // Is there going to be an issue here, having not accounted for an `r`?
+        match self {
+            Chunk::Numeric(n) => MChunk::Digits(*n, n.to_string()),
+            Chunk::Alphanum(s) => MChunk::Plain(s.clone()),
+        }
+        // match self.0.as_slice() {
+        //     [] => None,
+        //     [Unit::Digits(u)] => Some(MChunk::Digits(*u, u.to_string())),
+        //     [Unit::Letters(s), Unit::Digits(u)] if s == "r" => {
+        //         Some(MChunk::Rev(*u, format!("r{}", u)))
+        //     }
+        //     [Unit::Letters(s)] => Some(MChunk::Plain(s.clone())),
+        //     _ => Some(MChunk::Plain(format!("{}", self))),
+        // }
+    }
+
+    fn cmp_semver(&self, other: &Self) -> Ordering {
+        match (self, other) {
+            (Chunk::Numeric(a), Chunk::Numeric(b)) => a.cmp(&b),
+            (Chunk::Numeric(_), Chunk::Alphanum(_)) => Less,
+            (Chunk::Alphanum(_), Chunk::Numeric(_)) => Greater,
+            (Chunk::Alphanum(a), Chunk::Alphanum(b)) => a.cmp(&b),
+        }
+    }
+
+    fn cmp_lenient(&self, other: &Self) -> Ordering {
+        match (self, other) {
+            (Chunk::Numeric(a), Chunk::Numeric(b)) => a.cmp(&b),
+            (a @ Chunk::Alphanum(x), b @ Chunk::Alphanum(y)) => {
+                match (a.single_digit_lenient(), b.single_digit_lenient()) {
+                    (Some(i), Some(j)) => i.cmp(&j),
+                    _ => x.cmp(&y),
+                }
+            }
+            (Chunk::Numeric(n), b @ Chunk::Alphanum(_)) => match b.single_digit_lenient() {
+                None => Greater,
+                Some(m) => match n.cmp(&m) {
+                    // 1.2.0 > 1.2.0rc1
+                    Equal => Greater,
+                    c => c,
+                },
+            },
+            (a @ Chunk::Alphanum(_), Chunk::Numeric(n)) => match a.single_digit_lenient() {
+                None => Less,
+                Some(m) => match m.cmp(&n) {
+                    // 1.2.0rc1 < 1.2.0
+                    Equal => Less,
+                    c => c,
+                },
+            },
+        }
+    }
+}
+
+impl std::fmt::Display for Chunk {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        match self {
+            Chunk::Numeric(n) => write!(f, "{}", n),
+            Chunk::Alphanum(a) => write!(f, "{}", a),
+        }
     }
 }
 
@@ -1121,8 +1140,11 @@ impl fmt::Display for Chunks {
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[derive(Debug, PartialEq, Eq, Hash, Clone)]
 pub enum Versioning {
+    /// Follows good parsing and comparison rules.
     Ideal(SemVer),
+    /// A little more permissive than [`SemVer`].
     General(Version),
+    /// Hope that you need not venture here.
     Complex(Mess),
 }
 
@@ -1217,8 +1239,8 @@ impl Ord for Versioning {
     }
 }
 
-impl fmt::Display for Versioning {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+impl std::fmt::Display for Versioning {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         match self {
             Versioning::Ideal(s) => write!(f, "{}", s),
             Versioning::General(v) => write!(f, "{}", v),
@@ -1237,6 +1259,23 @@ impl Default for Versioning {
 mod tests {
     use super::*;
 
+    #[test]
+    fn chanks() {
+        assert_eq!(Ok(("", Chunk::Numeric(123))), Chunk::parse("123"));
+        assert_eq!(
+            Ok(("", Chunk::Alphanum("123a".to_string()))),
+            Chunk::parse("123a")
+        );
+        assert_eq!(
+            Ok(("", Chunk::Alphanum("123-456".to_string()))),
+            Chunk::parse("123-456")
+        );
+        assert_eq!(
+            Ok(("", Chunk::Alphanum("00a".to_string()))),
+            Chunk::parse("00a")
+        );
+    }
+
     #[test]
     fn official_semvers() {
         let goods = vec![
@@ -1279,17 +1318,10 @@ mod tests {
     }
 
     #[test]
-    fn semver_zeros() {
-        let pre_rel = Chunks(vec![Chunk(vec![
-            Unit::Letters("b".to_string()),
-            Unit::Digits(0),
-            Unit::Digits(5),
-        ])]);
+    fn tricky_semvers() {
+        let v = "1.2.2-00a";
 
-        assert_eq!(
-            Some(pre_rel),
-            SemVer::new("2.2.1-b05").and_then(|sv| sv.pre_rel)
-        );
+        assert_eq!("", SemVer::parse(v).unwrap().0);
     }
 
     #[test]
